// *********************************************************
//
// gradient.hh    : Gradient Application Include File
// author         : Fabio Silva and Chalermek Intanagonwiwat
//
// $Id: gradient.hh,v 1.12 2001/03/23 21:02:49 fabio Exp $
//
// *********************************************************

#ifndef GRADIENT_HH
#define GRADIENT_HH

#include <stdlib.h>
#include <stdio.h>
#include "dr.hh"
#include "utils.hh"

#define GRADIENT_APP_PRIORITY 2
#define OLD_MESSAGE -1
#define NEW_MESSAGE 1

typedef list<Tcl_HashEntry *> HashList;

class Hash_Entry {
public:
  int32_t last_hop;

  Hash_Entry() {
    last_hop = 0;
  }
};

class Attribute_Entry;
class Routing_Entry;
class Agents_Entry;

typedef list<Attribute_Entry *> AttributeList;
typedef list<Routing_Entry *> RoutingList;
typedef list<Agents_Entry *> AgentsList;

class Agents_Entry : public Agent_Entry {
public:
  bool reinforced;
  int neighbor_flag;

  Agents_Entry() : Agent_Entry() {
    reinforced = false;
    neighbor_flag = 0;
  };
};

class MyFilterReceive : public FilterCallback {
public:
  void recv(Message *msg, handle h);
};

class MyTimerReceive : public TimerCallbacks {
public:
  int recv(handle hdl, void *p);
  void del(void *p);
};

class TimerType {
public:
  int   which_timer;
  void  *param;

  TimerType(int _which_timer) : which_timer(_which_timer)
  {
    param = NULL;
  };

  ~TimerType() {};
};

class ReinforcementMessage {
public:
  int32_t rdm_id;
  int32_t pkt_num;
};

class Attribute_Entry {
public:
  NRAttrVec       *attrs;
  struct timeval  tv;

  Attribute_Entry() {
    attrs        = NULL;
    gettimeofday(&tv, NULL);
  };

  ~Attribute_Entry() {
    ClearAttrs(attrs);
    delete attrs;
  };
};

class Routing_Entry {
public:
  AgentsList      active;
  AgentsList      agents;
  NRAttrVec       *attrs;
  AttributeList   attr_list;
  struct timeval  tv;
  AgentsList      data_neighbors;
 
  Routing_Entry() {
    gettimeofday(&tv, NULL);
  };

  ~Routing_Entry() {
    AttributeList::iterator ATitr;
    AgentsList::iterator AGitr;
  
    // Clear Attributes
    ClearAttrs(attrs);
    delete attrs;

    ATitr = attr_list.begin();

    while (ATitr != attr_list.end()){
      delete (*ATitr);
      ATitr++;
    }
    attr_list.clear();

    AGitr = active.begin();

    while (AGitr != active.end()){
      delete (*AGitr);
      AGitr++;
    }
    active.clear();

    AGitr = agents.begin();

    while (AGitr != agents.end()){
      delete (*AGitr);
      AGitr++;
    }
    agents.clear();
  };
};

class GradientApp {
public:
  GradientApp(int argc, char **argv);
  void usage();
  void run();

  void recv(Message *msg, handle h);
  int ProcessTimers(handle hdl, void *p);

protected:

  // General Variables
  NR *dr;
  handle filterHandle;
  int pkt_count;
  int rdm_id;

  // Hashing structures
  HashList hash_list;
  Tcl_HashTable htable;

  // Receive Callback for the filter
  MyFilterReceive *fcb;
  MyTimerReceive *tcb;

  // List of all known datatypes
  RoutingList routing_list;

  // Setup the filter
  handle setupFilter();

  // Matching functions
  Routing_Entry * FindPerfectMatch(NRAttrVec *attrs);
  void DeleteRoutingEntry(Routing_Entry *entry);
  Routing_Entry * FindRoutingEntry(NRAttrVec *attrs, RoutingList::iterator start, RoutingList::iterator *place);
  Attribute_Entry * FindMatchingSubscription(Routing_Entry *entry, NRAttrVec *attrs);

  // Data structure management
  void UpdateGradient(Routing_Entry *entry, int32_t last_hop, bool reinforced);
  void UpdateAgent(Routing_Entry *entry, u_int16_t source_port);
  Agents_Entry * FindReinforcedGradients(AgentsList *agents, AgentsList::iterator start, AgentsList::iterator *place);
  Agents_Entry * FindReinforcedGradient(int32_t node_addr, Routing_Entry *entry);
  void DeleteGradient(Routing_Entry *entry, Agents_Entry *ae);
  void SetReinforcementFlags(Routing_Entry *entry, int32_t last_hop, int new_message);

  // Message forwarding functions
  void SendInterest(NRAttrVec *attrs, Routing_Entry *entry);
  void SendDisinterest(NRAttrVec *attrs, Routing_Entry *entry);
  void ForwardData(Message *msg, Routing_Entry *entry, bool &reinforced);

  // Message Processing functions
  void ProcessOldMessage(Message *msg);
  void ProcessNewMessage(Message *msg);

  // Timers
  void DataTimeout(Message *msg);
  void InterestTimeout(Message *msg);
  void GradientTimeout();
  void ReinforcementTimeout();
  int SubscriptionTimeout(NRAttrVec *attrs);
  void TimetoStop();
  
  // Hashing functions
  Hash_Entry *GetHash(unsigned int, unsigned int);
  void PutHash(Hash_Entry *newHashPtr, unsigned int, unsigned int);
};

#endif //GRADIENT_HH
