// *********************************************************
//
// dr.cc           : Diffusion Routing Class
// authors         : John Heidemann and Fabio Silva
//
// $Id: dr.cc,v 1.29 2001/04/05 21:28:26 fabio Exp $
//
// *********************************************************

#include <stdlib.h>
#include <stdio.h>
#include "dr.hh"

NR *dr = NULL;

void * ReceiveThread(void *dr)
{
  // Never returns
  ((DiffusionRouting *)dr)->run();

  return NULL;
}

NR * NR::createNR()
{
  int retval;
  pthread_t thread;

  // Create Diffusion Routing Class
  if (dr)
    return dr;

  dr = new DiffusionRouting;

  // Fork a thread for receiving Messages
  retval = pthread_create(&thread, NULL, &ReceiveThread, (void *)dr);

  if (retval){
    fprintf(stderr, "Error creating receiving thread ! Aborting...\n");
    exit(-1);
  }

  return dr;
}

DiffusionRouting::DiffusionRouting()
{
  struct sockaddr_in my_addr_inet;
  socklen_t my_addr_inet_size = sizeof(my_addr_inet);

  // Initialize basic stuff
  next_handle = 1;
  srand(time(NULL));
  pkt_count = rand();
  rdm_id = rand();

  // Initialize event queue
  eq = new eventQueue;
  eq->eq_new();

  // Open sockets
  host_ent = GetThisHostEnt();
  in_sock = CreateSocket(host_ent, 0);
  out_sock = CreateSocket(host_ent, DIFFUSION_PORT);
  RegisterSocket(in_sock);

  if (getsockname(in_sock->skt_fd, (struct sockaddr *) &(my_addr_inet),
		  &my_addr_inet_size) != 0){
    fprintf(stderr, "Failed to open incoming socket !\n");
    exit(-1);
  }

  agent_id = ntohs(my_addr_inet.sin_port);

  fprintf(stderr, "Diffusion Routing Agent: Initializing... Agent ID = %d\n",
	  agent_id);
  listening = true;

  // Set pointer to packet headers
  in_packet_dfh = HDR_DIFF(in_buffer);
  out_dfh = HDR_DIFF(out_pkt);
  addrlen = SZ_ADDR_INET;

  // Initialize Semaphores
  drMtx = new pthread_mutex_t;
  queueMtx = new pthread_mutex_t;

  pthread_mutex_init(drMtx, NULL);
  pthread_mutex_init(queueMtx, NULL);
}

handle DiffusionRouting::subscribe(NRAttrVec *subscribeAttrs, NR::Callback *cb)
{
  Handle_Entry *my_handle;
  NRAttribute *scopeAttr;

  // Get lock first
  pthread_mutex_lock(drMtx);

  // Check the published attributes
  if (!checkSubscription(subscribeAttrs)){
    fprintf(stderr, "Error : Invalid class/scope attributes in the subscribe attributes !\n");
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  // Create and Initialize the handle_entry structute
  my_handle = new Handle_Entry;
  my_handle->hdl = next_handle;
  next_handle++;
  my_handle->cb = (NR::Callback *) cb;
  sub_list.push_back(my_handle);

  // Copy the attributes   
  my_handle->attrs = CopyAttrs(subscribeAttrs);

  // For subscriptions, scope is global if not specified
  if (!hasScope(subscribeAttrs)){
    scopeAttr = NRScopeAttr.make(NRAttribute::IS, NRAttribute::GLOBAL_SCOPE);
    my_handle->attrs->push_back(scopeAttr);
  }

  // Send interest and add it to the queue
  InterestTimeout(my_handle);

  // Release lock
  pthread_mutex_unlock(drMtx);

  return my_handle->hdl;
}

int DiffusionRouting::unsubscribe(handle subscription_handle)
{
  Handle_Entry *my_handle = NULL;

  // Get the lock first
  pthread_mutex_lock(drMtx);

  my_handle = findHandle(subscription_handle, &sub_list);
  if (!my_handle){
    // Handle doesn't exist, return FAIL
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  // Handle will be destroyed when next InterestTimeout happens
  my_handle->valid = false;

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return OK;
}

handle DiffusionRouting::publish(NRAttrVec *publishAttrs)
{
  Handle_Entry *my_handle;
  NRAttribute *scopeAttr;

  // Get the lock first
  pthread_mutex_lock(drMtx);

  // Check the published attributes
  if (!checkPublication(publishAttrs)){
    fprintf(stderr, "Error : Invalid class/scope attributes in the publish attributes !\n");
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  // Create and Initialize the handle_entry structute
  my_handle = new Handle_Entry;
  my_handle->hdl = next_handle;
  next_handle++;
  pub_list.push_back(my_handle);

  // Copy the attributes
  my_handle->attrs = CopyAttrs(publishAttrs);

  // For publications, scope is local if not specified
  if (!hasScope(publishAttrs)){
    scopeAttr = NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE);
    my_handle->attrs->push_back(scopeAttr);
  }

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return my_handle->hdl;
}

int DiffusionRouting::unpublish(handle publication_handle)
{
  Handle_Entry *my_handle = NULL;

  // Get the lock first
  pthread_mutex_lock(drMtx);

  my_handle = removeHandle(publication_handle, &pub_list);
  if (!my_handle){
    // Handle doesn't exist, return FAIL
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  // Free structures
  delete my_handle;

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return OK;
}

int DiffusionRouting::send(handle publication_handle,
			   NRAttrVec *sendAttrs)
{
  Handle_Entry *my_handle;
  Packed_Attribute *aux;
  char *pos;
  int len;

  // Get the lock first
  pthread_mutex_lock(drMtx);

  // Get attributes associated with handle
  my_handle = findHandle(publication_handle, &pub_list);
  if (!my_handle){
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  // Check the send attributes
  if (!checkSend(sendAttrs)){
    fprintf(stderr, "Error : Invalid class/scope attributes in the send attributes !\n");
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  pos = (char *) &out_pkt[0];
  pos = pos + sizeof(struct hdr_diff);

  len = PackAttrs(my_handle->attrs, pos);

  // Move pointer ahead
  pos = pos + len;

  // Add the attributes passed in send
  len = len + PackAttrs(sendAttrs, pos);

  // Prepare packet header to send
  LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_dfh) = htons(sendAttrs->size() + my_handle->attrs->size());
  PKT_NUM(out_dfh) = htonl(pkt_count);
  RDM_ID(out_dfh) = htonl(rdm_id);
  pkt_count++;
  SRC_PORT(out_dfh) = htons(agent_id);
  VERSION(out_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_dfh) = DATA;
  DATA_LEN(out_dfh) = htons(len);

  // Send Packet
  SEND_TO(out_sock, out_pkt);

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return OK;
}

handle DiffusionRouting::addFilter(NRAttrVec *filterAttrs, u_int16_t priority,
				   FilterCallback *cb)
{
  Filter_Entry *my_filter;
  NRAttrVec *attrs;
  NRAttribute *ctrlmsg;
  ControlMessage *controlblob;
  int len;
  char *pos;

  // Get lock first
  pthread_mutex_lock(drMtx);

  // Create and Initialize the handle_entry structute
  my_filter = new Filter_Entry(next_handle, priority, agent_id);
  next_handle++;
  my_filter->cb = (FilterCallback *) cb;
  filter_list.push_back(my_filter);

  // Copy attributes (keep them for matching later)
  my_filter->filterAttrs = CopyAttrs(filterAttrs);

  // Copy the attributes (and add the control attr)
  attrs = CopyAttrs(filterAttrs);
  controlblob = new ControlMessage;
  controlblob->command = ADD_FILTER;
  controlblob->param1 = priority;
  controlblob->param2 = my_filter->handle;

  ctrlmsg = ControlMsgAttr.make(NRAttribute::IS, (void *)controlblob, sizeof(ControlMessage));

  attrs->push_back(ctrlmsg);

  pos = (char *) &out_pkt[0];
  pos = pos + sizeof(struct hdr_diff);

  len = PackAttrs(attrs, pos);

  // Prepare packet header to send
  LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_dfh) = htons(attrs->size());
  PKT_NUM(out_dfh) = htonl(pkt_count);
  pkt_count++;
  RDM_ID(out_dfh) = htonl(rdm_id);
  SRC_PORT(out_dfh) = htons(agent_id);
  VERSION(out_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_dfh) = CONTROL;
  DATA_LEN(out_dfh) = htons(len);

  // Send Packet
  SEND_TO(out_sock, out_pkt);

  // Add keepalive to event queue
  pthread_mutex_lock(queueMtx);
  eq->eq_addAfter(FILTER_KEEPALIVE_TIMER, (void *) my_filter,
		  FILTER_KEEPALIVE_DELAY);
  pthread_mutex_unlock(queueMtx);

  ClearAttrs(attrs);
  delete attrs;
  delete controlblob;

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return my_filter->handle;
}

int DiffusionRouting::removeFilter(handle filterHandle)
{
  Filter_Entry *entry = NULL;
  ControlMessage *controlblob;
  NRAttribute *ctrlmsg;
  NRAttrVec attrs;
  int len;
  char *pos;

  // Get lock first
  pthread_mutex_lock(drMtx);

  entry = findFilter(filterHandle);
  if (!entry){
    // Handle doesn't exist, return FAIL
    pthread_mutex_unlock(drMtx);
    return FAIL;
  }

  pos = (char *) &out_pkt[0];
  pos = pos + sizeof(struct hdr_diff);

  controlblob = new ControlMessage;
  controlblob->command = REMOVE_FILTER;
  controlblob->param1 = entry->handle;
  controlblob->param2 = 0;

  ctrlmsg = ControlMsgAttr.make(NRAttribute::IS, (void *)controlblob, sizeof(ControlMessage));

  attrs.push_back(ctrlmsg);

  len = PackAttrs(&attrs, pos);

  // Prepare packet header to send
  LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_dfh) = htons(attrs.size());
  PKT_NUM(out_dfh) = htonl(pkt_count);
  pkt_count++;
  RDM_ID(out_dfh) = htonl(rdm_id);
  SRC_PORT(out_dfh) = htons(agent_id);
  VERSION(out_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_dfh) = CONTROL;
  DATA_LEN(out_dfh) = htons(len);

  // Send Packet
  SEND_TO(out_sock, out_pkt);

  // Handle will be destroyed when next keepalive timer happens
  entry->valid = false;

  ClearAttrs(&attrs);
  delete controlblob;

  // Release the lock
  pthread_mutex_unlock(drMtx);

  return OK;
}

handle DiffusionRouting::addTimer(int timeout, void *p, TimerCallbacks *cb)
{
  Timer_Entry *entry;

  entry = new Timer_Entry(next_handle, timeout, p, cb);

  pthread_mutex_lock(queueMtx);
  eq->eq_addAfter(APPLICATION_TIMER, (void *) entry, timeout);
  pthread_mutex_unlock(queueMtx);

  next_handle++;

  return entry->hdl;
}

int DiffusionRouting::removeTimer(handle hdl)
{
  event *e;
  Timer_Entry *entry;
  int found = -1;

  pthread_mutex_lock(queueMtx);

  // Find the timer in the queue
  e = eq->eq_findEvent(APPLICATION_TIMER);
  while (e){
    entry = (Timer_Entry *) e->payload;
    if (entry->hdl == hdl){
      found = 0;
      break;
    }

    e = eq->eq_findNextEvent(APPLICATION_TIMER, e->next);
  }

  // If timer found, remove it from the queue
  if (e){
    if (eq->eq_remove(e) != 0){
      fprintf(stderr, "Error: Can't remove event from queue !\n");
      abort();
    }

    // Call the application provided delete function
    entry->cb->del(entry->p);

    delete entry;
    free(e);
  }

  pthread_mutex_unlock(queueMtx);

  return found;
}

void DiffusionRouting::FilterKeepaliveTimeout(Filter_Entry *entry)
{
  Filter_Entry *my_handle = NULL;
  ControlMessage *controlblob;
  NRAttribute *ctrlmsg;
  NRAttrVec attrs;
  char *pos;
  int len;

  if (entry->valid){
    // Send keepalive

    pos = (char *) &out_pkt[0];
    pos = pos + sizeof(struct hdr_diff);

    controlblob = new ControlMessage;
    controlblob->command = KEEPALIVE_FILTER;
    controlblob->param1 = entry->handle;
    controlblob->param2 = 0;

    ctrlmsg = ControlMsgAttr.make(NRAttribute::IS, (void *)controlblob, sizeof(ControlMessage));

    attrs.push_back(ctrlmsg);

    len = PackAttrs(&attrs, pos);

    // Prepare packet header to send
    LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
    NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
    NUM_ATTR(out_dfh) = htons(attrs.size());
    PKT_NUM(out_dfh) = htonl(pkt_count);
    pkt_count++;
    RDM_ID(out_dfh) = htonl(rdm_id);
    SRC_PORT(out_dfh) = htons(agent_id);
    VERSION(out_dfh) = DIFFUSION_VERSION;
    MSG_TYPE(out_dfh) = CONTROL;
    DATA_LEN(out_dfh) = htons(len);

    // Send Packet
    SEND_TO(out_sock, out_pkt);

    // Add another keepalive to event queue
    pthread_mutex_lock(queueMtx);
    eq->eq_addAfter(FILTER_KEEPALIVE_TIMER, (void *) entry,
		    FILTER_KEEPALIVE_DELAY);
    pthread_mutex_unlock(queueMtx);

    ClearAttrs(&attrs);
    delete controlblob;
  }
  else{
    // Filter was removed
    my_handle = deleteFilter(entry->handle);

    // We should have removed the correct handle
    if (my_handle != entry){
      fprintf(stderr, "DiffusionRouting::KeepaliveTimeout: Handles should match !\n");
      exit(-1);
    }

    delete my_handle;
  }
}

void DiffusionRouting::InterestTimeout(Handle_Entry *entry)
{
  Handle_Entry *my_handle = NULL;
  char *pos;
  int len;

  if (entry->valid){
    // Send the interest

    pos = (char *) &out_pkt[0];
    pos = pos + sizeof(struct hdr_diff);

    len = PackAttrs(entry->attrs, pos);

    // Prepare packet header to send
    LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
    NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
    NUM_ATTR(out_dfh) = htons(entry->attrs->size());
    PKT_NUM(out_dfh) = htonl(pkt_count);
    pkt_count++;
    RDM_ID(out_dfh) = htonl(rdm_id);
    SRC_PORT(out_dfh) = htons(agent_id);
    VERSION(out_dfh) = DIFFUSION_VERSION;
    MSG_TYPE(out_dfh) = INTEREST;
    DATA_LEN(out_dfh) = htons(len);

    // Send Packet
    SEND_TO(out_sock, out_pkt);

    // Add another InterestTimeout to the queue
    pthread_mutex_lock(queueMtx);
    eq->eq_addAfter(INTEREST_TIMER, (void *) entry,
		    INTEREST_DELAY +
		    (int) ((INTEREST_JITTER * (rand() * 1.0 / RAND_MAX)) -
			   (INTEREST_JITTER / 2)));
    pthread_mutex_unlock(queueMtx);
  }
  else{
    // Interest was canceled. Just delete it from the handle_list
    my_handle = removeHandle(entry->hdl, &sub_list);

    // We should have removed the correct handle
    if (my_handle != entry){
      fprintf(stderr, "DiffusionRouting::InterestTimeout: Handles should match !\n");
      exit(-1);
    }

    delete my_handle;
  }
}

void DiffusionRouting::sendMessage(Message *msg, handle h,
				   u_int16_t dst_agent_id = 0)
{
  RedirectMessage *originalHdr;
  NRAttribute *originalAttr, *ctrlmsg;
  ControlMessage *controlblob;
  NRAttrVec *attrs;
  int len;
  char *pos;

  // Create an attribute with the original header
  originalHdr = new RedirectMessage;
  originalHdr->msg_type = msg->msg_type;
  originalHdr->source_port = msg->source_port;
  originalHdr->data_len = msg->data_len;
  originalHdr->num_attr = msg->num_attr;
  originalHdr->pkt_num = msg->pkt_num;
  originalHdr->rdm_id = msg->rdm_id;
  originalHdr->next_hop = msg->next_hop;
  originalHdr->last_hop = msg->last_hop;
  originalHdr->new_message = msg->new_message;
  originalHdr->handle = 0;

  originalAttr = OriginalHdrAttr.make(NRAttribute::IS, (void *)originalHdr, sizeof(RedirectMessage));

  // Create the attribute with the control message
  controlblob = new ControlMessage;
  controlblob->command = SEND_MESSAGE;
  controlblob->param1 = dst_agent_id;
  controlblob->param2 = 0;

  ctrlmsg = ControlMsgAttr.make(NRAttribute::IS, (void *)controlblob, sizeof(ControlMessage));

  // Copy Attributes and add the originalAttr
  attrs = CopyAttrs(msg->msg_attr_vec);
  attrs->push_back(originalAttr);
  attrs->push_back(ctrlmsg);

  // Send the CONTROL message to diffusion
  pos = (char *) &out_pkt[0];
  pos = pos + sizeof(struct hdr_diff);

  len = PackAttrs(attrs, pos);

  // Prepare packet header to send
  LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_dfh) = htons(attrs->size());
  PKT_NUM(out_dfh) = htonl(pkt_count);
  pkt_count++;
  RDM_ID(out_dfh) = htonl(rdm_id);
  SRC_PORT(out_dfh) = htons(agent_id);
  VERSION(out_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_dfh) = CONTROL;
  DATA_LEN(out_dfh) = htons(len);

  // Send Packet
  SEND_TO(out_sock, out_pkt);

  ClearAttrs(attrs);
  delete attrs;
  delete controlblob;
  delete originalHdr;
}

void DiffusionRouting::sendMessageToNext(Message *msg, handle h)
{
  RedirectMessage *originalHdr;
  NRAttribute *originalAttr, *ctrlmsg;
  ControlMessage *controlblob;
  NRAttrVec *attrs;
  int len;
  char *pos;

  // Create an attribute with the original header
  originalHdr = new RedirectMessage;
  originalHdr->msg_type = msg->msg_type;
  originalHdr->source_port = msg->source_port;
  originalHdr->data_len = msg->data_len;
  originalHdr->num_attr = msg->num_attr;
  originalHdr->pkt_num = msg->pkt_num;
  originalHdr->rdm_id = msg->rdm_id;
  originalHdr->next_hop = msg->next_hop;
  originalHdr->last_hop = msg->last_hop;
  originalHdr->new_message = msg->new_message;
  originalHdr->handle = 0;

  originalAttr = OriginalHdrAttr.make(NRAttribute::IS, (void *)originalHdr, sizeof(RedirectMessage));

  // Create the attribute with the control message
  controlblob = new ControlMessage;
  controlblob->command = SEND_TO_NEXT;
  controlblob->param1 = h;
  controlblob->param2 = 0;

  ctrlmsg = ControlMsgAttr.make(NRAttribute::IS, (void *)controlblob, sizeof(ControlMessage));

  // Copy Attributes and add originalAttr and controlAttr
  attrs = CopyAttrs(msg->msg_attr_vec);
  attrs->push_back(originalAttr);
  attrs->push_back(ctrlmsg);

  // Send the CONTROL message to diffusion
  pos = (char *) &out_pkt[0];
  pos = pos + sizeof(struct hdr_diff);

  len = PackAttrs(attrs, pos);

  // Prepare packet header to send
  LAST_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NEXT_HOP(out_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_dfh) = htons(attrs->size());
  PKT_NUM(out_dfh) = htonl(pkt_count);
  pkt_count++;
  RDM_ID(out_dfh) = htonl(rdm_id);
  SRC_PORT(out_dfh) = htons(agent_id);
  VERSION(out_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_dfh) = CONTROL;
  DATA_LEN(out_dfh) = htons(len);

  // Send Packet
  SEND_TO(out_sock, out_pkt);

  ClearAttrs(attrs);
  delete attrs;
  delete controlblob;
  delete originalHdr;
}

void DiffusionRouting::run()
{
  int status;
  fd_set fds;
  struct timeval *tv;
  struct timeval tmv;

  while (listening){
    FD_ZERO(&fds);
    FD_SET(in_sock->skt_fd, &fds);

    // Get the next timeout
    pthread_mutex_lock(queueMtx);
    tv = eq->eq_nextTimer();
    pthread_mutex_unlock(queueMtx);

    // If we don't have any timeouts, we wait for POLLING_INTERVAL
    if (!tv){
      tv = &tmv;
      tmv.tv_sec = POLLING_INTERVAL;
      tmv.tv_usec = 0;
    }

    fflush(stdout);
    status = select((in_sock->skt_fd)+1, &fds, NULL, NULL, tv);

    pthread_mutex_lock(queueMtx);
    if (eq->eq_topInPast()){
      // We got a timeout
      event *e = eq->eq_pop();
      pthread_mutex_unlock(queueMtx);

      // Timeouts
      switch (e->type){

      case INTEREST_TIMER:

	pthread_mutex_lock(drMtx);
	InterestTimeout((Handle_Entry *) e->payload);
	pthread_mutex_unlock(drMtx);

	free(e);

	break;

      case FILTER_KEEPALIVE_TIMER:

	pthread_mutex_lock(drMtx);
	FilterKeepaliveTimeout((Filter_Entry *) e->payload);
	pthread_mutex_unlock(drMtx);

	free(e);

	break;

      case APPLICATION_TIMER:

	Timer_Entry *entry;
	int new_timeout;

	entry = (Timer_Entry *) e->payload;

	new_timeout = entry->cb->recv(entry->hdl, entry->p);

	pthread_mutex_lock(queueMtx);
	if (new_timeout >= 0){
	  if (new_timeout > 0){
	    // Change the timer's timeout
	    entry->timeout = new_timeout;
	  }

	  eq->eq_addAfter(APPLICATION_TIMER, (void *) entry, entry->timeout);
	}
	else{
	  entry->cb->del(entry->p);
	  delete entry;
	}

	pthread_mutex_unlock(queueMtx);
	free(e);

	break;

      default:

	break;

      }
    }
    else{
      // Don't forget to release the lock
      pthread_mutex_unlock(queueMtx);

      if (status > 0){
	RECV_FROM(in_sock, in_buffer, addrlen);
	recv();
      }
      else
	if (status < 0)
	  fprintf(stderr, "Select returned %d\n", status);
    }
  }
}

void DiffusionRouting::recv()
{
  Message *rcv_message = NULL;
  int8_t version, msg_type;
  u_int16_t data_len, num_attr, source_port;
  int32_t pkt_num, rdm_id, next_hop, last_hop;

  // Read header
  version = VERSION(in_packet_dfh);
  msg_type = MSG_TYPE(in_packet_dfh);
  source_port = ntohs(SRC_PORT(in_packet_dfh));
  pkt_num = ntohl(PKT_NUM(in_packet_dfh));
  rdm_id = ntohl(RDM_ID(in_packet_dfh));
  num_attr = ntohs(NUM_ATTR(in_packet_dfh));
  next_hop = ntohl(NEXT_HOP(in_packet_dfh));
  last_hop = ntohl(LAST_HOP(in_packet_dfh));
  data_len = ntohs(DATA_LEN(in_packet_dfh));

  // Check version
  if (version != DIFFUSION_VERSION)
    return;

  // Check destination
  if (next_hop != LOCALHOST_ADDR)
    return;

  // Packet is good, create a message
  rcv_message = new Message(version, msg_type, source_port, data_len,
			    num_attr, pkt_num, rdm_id, next_hop, last_hop);

  int total_size = sizeof(struct hdr_diff) + data_len;
  rcv_message->msg = new int [1 + (total_size / sizeof(int))];
  if (rcv_message->msg == NULL){
    fprintf(stderr, "Can't allocate memory. Duplicate Packet fails.\n");
    exit(-1);
  }

  bzero(rcv_message->msg, total_size);
  memcpy(rcv_message->msg, in_buffer, total_size);
  rcv_message->msg_size = total_size;

  // Read all attributes into the Message structure
  rcv_message->msg_attr_vec = UnpackAttrs(rcv_message->msg, num_attr);

  // Process the incoming message
  if (rcv_message->msg_type == REDIRECT)
    ProcessControlMessage(rcv_message);
  else
    ProcessMessage(rcv_message);

  // We are dome
  delete rcv_message;
}

void DiffusionRouting::ProcessControlMessage(Message *msg)
{
  NRSimpleAttribute<void *> *originalHeader = NULL;
  NRAttrVec::iterator place = msg->msg_attr_vec->begin();
  RedirectMessage *originalHdr;
  Filter_Entry *entry;
  handle my_handle;

  // Find the attribute containing the original packet header
  originalHeader = OriginalHdrAttr.find_from(msg->msg_attr_vec, place, &place);
  if (!originalHeader){
    fprintf(stderr, "Error: DiffusionRouting::ProcessControlMessage: Received an invalid REDIRECT message !\n");
    return;
  }

  // Restore original message header
  originalHdr = (RedirectMessage *) originalHeader->getVal();
  my_handle = originalHdr->handle;
  msg->msg_type = originalHdr->msg_type;
  msg->source_port = originalHdr->source_port;
  msg->pkt_num = originalHdr->pkt_num;
  msg->rdm_id = originalHdr->rdm_id;
  msg->next_hop = originalHdr->next_hop;
  msg->last_hop = originalHdr->last_hop;
  msg->num_attr = originalHdr->num_attr;
  msg->new_message = originalHdr->new_message;

  // Delete attribute from the original set
  msg->msg_attr_vec->erase(place);
  delete originalHeader;

  // Find the right callback
  pthread_mutex_lock(drMtx);

  entry = findFilter(my_handle);
  if (entry && entry->valid){
    // Just to confirm
    if (OneWayMatch(entry->filterAttrs, msg->msg_attr_vec)){
      entry->cb->recv(msg, my_handle);
    }
    else{
      fprintf(stderr, "Warning: DiffusionRouting::ProcessControlMessage: Filter specified doesn't match incoming message's attributes !\n");
    }
  }
  else{
    fprintf(stderr, "Report: DiffusionRouting::ProcessControlMessage: Filter in REDIRECT message was not found (possibly deleted ?)\n");
  }

  pthread_mutex_unlock(drMtx);
}

void DiffusionRouting::ProcessMessage(Message *msg)
{
  CallbackList cbl;
  Callback_Entry *aux;
  HandleList::iterator sub_itr;
  CallbackList::iterator cbl_itr;
  Handle_Entry *entry; 
  NRAttrVec *callbackAttrs;

  // Now, acquire the lock
  pthread_mutex_lock(drMtx);

  for (sub_itr = sub_list.begin(); sub_itr != sub_list.end(); ++sub_itr){
    entry = *sub_itr;
    if ((entry->valid) && (MatchAttrs(msg->msg_attr_vec, entry->attrs)))
      if (entry->cb){
	aux = new Callback_Entry;
	aux->cb = entry->cb;
	aux->subscription_handle = entry->hdl;
	cbl.push_back(aux);
      }
  }

  // We can release the lock now
  pthread_mutex_unlock(drMtx);

  // Now we just call all callback functions
  for (cbl_itr = cbl.begin(); cbl_itr != cbl.end(); ++cbl_itr){
    // Copy attributes
    callbackAttrs = CopyAttrs(msg->msg_attr_vec);

    // Call app-specific callback function
    aux = *cbl_itr;
    aux->cb->recv(msg->msg_attr_vec, aux->subscription_handle);
    delete aux;

    // Clean up callback attributes
    ClearAttrs(callbackAttrs);
    delete callbackAttrs;
  }

  // We are done
  cbl.clear();
}

Handle_Entry * DiffusionRouting::removeHandle(handle my_handle, HandleList *hl)
{
  HandleList::iterator itr;
  Handle_Entry *entry;

  for (itr = hl->begin(); itr != hl->end(); ++itr){
    entry = *itr;
    if (entry->hdl == my_handle){
      hl->erase(itr);
      return entry;
    }
  }
  return NULL;
}

Handle_Entry * DiffusionRouting::findHandle(handle my_handle, HandleList *hl)
{
  HandleList::iterator itr;
  Handle_Entry *entry;

  for (itr = hl->begin(); itr != hl->end(); ++itr){
    entry = *itr;
    if (entry->hdl == my_handle)
      return entry;
  }
  return NULL;
}

Filter_Entry * DiffusionRouting::deleteFilter(handle my_handle)
{
  FilterList::iterator itr;
  Filter_Entry *entry;

  for (itr = filter_list.begin(); itr != filter_list.end(); ++itr){
    entry = *itr;
    if (entry->handle == my_handle){
      filter_list.erase(itr);
      return entry;
    }
  }
  return NULL;
}

Filter_Entry * DiffusionRouting::findFilter(handle my_handle)
{
  FilterList::iterator itr;
  Filter_Entry *entry;

  for (itr = filter_list.begin(); itr != filter_list.end(); ++itr){
    entry = *itr;
    if (entry->handle == my_handle)
      return entry;
  }
  return NULL;
}

bool DiffusionRouting::hasScope(NRAttrVec *attrs)
{
  NRAttribute *temp = NULL;

  temp = NRScopeAttr.find(attrs);
  if (temp)
    return true;

  return false;
}

bool DiffusionRouting::checkSubscription(NRAttrVec *attrs)
{
  NRSimpleAttribute<int> *nrclass = NULL;
  NRSimpleAttribute<int> *nrscope = NULL;

  // Checks for class attribute and possible scope attribute
  nrclass = NRClassAttr.find(attrs);
  nrscope = NRScopeAttr.find(attrs);

  if (!nrclass)
    return false;

  if (nrscope){
    // Check for global interests
    if ((nrscope->getVal() == NRAttribute::GLOBAL_SCOPE) &&
	(nrclass->getVal() == NRAttribute::INTEREST_CLASS) &&
	(nrclass->getOp() == NRAttribute::IS))
      return true;

    // Check for local subscriptions
    if ((nrscope->getVal() == NRAttribute::NODE_LOCAL_SCOPE) &&
	(nrclass->getOp() != NRAttribute::IS))
      return true;
  }

  if ((nrclass->getVal() == NRAttribute::INTEREST_CLASS) &&
      (nrclass->getOp() == NRAttribute::IS))
    return true;

  return false;
}

bool DiffusionRouting::checkPublication(NRAttrVec *attrs)
{
  NRSimpleAttribute<int> *nrclass = NULL;
  NRSimpleAttribute<int> *nrscope = NULL;

  // Checks for class attribute and possible scope attribute
  nrclass = NRClassAttr.find(attrs);
  nrscope = NRScopeAttr.find(attrs);

  if (!nrclass)
    return false;

  if ((nrclass->getVal() != NRAttribute::DATA_CLASS) ||
      (nrclass->getOp() != NRAttribute::IS))
    return false;

  // No scope attribute is OK
  if (!nrscope)
    return true;

  if ((nrscope->getOp() != NRAttribute::IS) ||
      (nrscope->getVal() != NRAttribute::NODE_LOCAL_SCOPE))
    return false;

  return true;
}

bool DiffusionRouting::checkSend(NRAttrVec *attrs)
{
  NRSimpleAttribute<int> *nrclass = NULL;
  NRSimpleAttribute<int> *nrscope = NULL;

  // Currently only checks for Class and Scope attributes
  nrclass = NRClassAttr.find(attrs);
  nrscope = NRScopeAttr.find(attrs);

  if (nrclass || nrscope)
    return false;

  return true;
}
