// *********************************************************
// 
// diffusion.cc  : Main Diffusion program
// authors       : Chalermek Intanagonwiwat and Fabio Silva
//
// $Id: diffusion.cc,v 1.43 2001/05/22 22:13:30 fabio Exp $
//
// *********************************************************

extern "C" {
#include <fcntl.h>

#ifndef WIRED
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif
}

#include <signal.h>

#include "diffusion.hh"

DiffusionAgent *agent;

void signal_handler(int p)
{
  agent->TimetoStop();
  exit(-1);
}

void DiffusionAgent::NeighborsTimeOut()
{
  struct timeval tmv;
  Neighbor_Entry *myNeighbor;
  NeighborList::iterator itr;

  fprintf(stderr, "Neighbors Timeout !\n");

  gettimeofday(&tmv, NULL);

  itr = my_neighbors.begin();

  while(itr != my_neighbors.end()){
    myNeighbor = *itr;
    if (tmv.tv_sec > myNeighbor->tmv.tv_sec + NEIGHBORS_TIMEOUT){
      // This neighbor expired
      itr = my_neighbors.erase(itr);
      delete myNeighbor;
    }
    else{
      itr++;
    }
  }

  // Re-schedule this timer
  eq->eq_addAfter(NEIGHBORS_TIMER, NULL, NEIGHBORS_DELAY);
}

void DiffusionAgent::FilterTimeOut()
{
  struct timeval tmv;
  Filter_Entry *myFilter;
  FilterList::iterator itr;

  fprintf(stderr, "Filter Timeout !\n");

  gettimeofday(&tmv, NULL);

  itr = filter_list.begin();

  while(itr != filter_list.end()){
    myFilter = *itr;
    if (tmv.tv_sec > myFilter->tmv.tv_sec + FILTER_TIMEOUT){
      // This filter expired
      fprintf(stderr, "Filter %d, %d, %d timed out !\n", myFilter->agent,
	      myFilter->handle, myFilter->priority);
      itr = filter_list.erase(itr);
      delete myFilter;
    }
    else{
      itr++;
    }
  }

  // Re-Schedule this timer
  eq->eq_addAfter(FILTER_TIMER, NULL, FILTER_DELAY);
}

void DiffusionAgent::SendMessage(Message *msg, u_int16_t port)
{
  Tcl_HashEntry *entryPtr;
  unsigned int key[2];
  int len;
  char *pos;
  
  pos = (char *) &out_buffer;
  pos = pos + sizeof(struct hdr_diff);

  // Clear the buffer
  len = CalculateSize(msg->msg_attr_vec);
  bzero(&out_buffer, (len + sizeof(struct hdr_diff)));
  len = PackAttrs(msg->msg_attr_vec, pos);

  // Adjust message size for logging and check hash
  msg->data_len = len;
  key[0] = msg->pkt_num;
  key[1] = msg->rdm_id;
  entryPtr = Tcl_FindHashEntry(&htable, (char *) key);
  if (entryPtr)
    msg->new_message = 0;
  else
    msg->new_message = 1;

  NEXT_HOP(out_packet_dfh) = htonl(msg->next_hop);
  NUM_ATTR(out_packet_dfh) = htons(msg->msg_attr_vec->size());
  SRC_PORT(out_packet_dfh) = htons(DIFFUSION_PORT);
  MSG_TYPE(out_packet_dfh) = msg->msg_type;
  VERSION(out_packet_dfh)  = DIFFUSION_VERSION;
  DATA_LEN(out_packet_dfh) = htons(len);
  PKT_NUM(out_packet_dfh)  = htonl(msg->pkt_num);
  RDM_ID(out_packet_dfh)   = htonl(msg->rdm_id);

  // Check if message goes to an agent or the network
  if (port){
    // Message goes to an agent
    LAST_HOP(out_packet_dfh) = htonl(LOCALHOST_ADDR);

    if (msg->next_hop != LOCALHOST_ADDR){
      fprintf(stderr, "Error: DiffusionAgent::SendMessage: message destination is a local agent but next_hop != LOCALHOST_ADDR !\n");
      return;
    }

    // Find the agent's socket
    AgentList::iterator itr;
    Agent_Entry *entry;

    for (itr = agent_list.begin(); itr != agent_list.end(); ++itr){
      entry = *itr;
      if (entry->port == port){
	// Agent Found
	SEND_TO(entry->skt_info, out_buffer);
	return;
      }
    }

    // We shouldn't reach this point
    fprintf(stderr, "Error: DiffusionAgent::SendMessage: message destination agent can't be foind !\n");
  }
  else{
    // Message goes to the network
    LAST_HOP(out_packet_dfh) = htonl(myId);

#ifdef STATS
    stats->logOutgoingMessage(msg);
#endif

    // Add message to the hash table      
    if (entryPtr == NULL)
      PutHash(key[0], key[1]);
    else
      fprintf(stderr, "Message being sent is an old message !\n");

    // Send Message
    dvi_send((char *) out_buffer, sizeof(struct hdr_diff) + len);
  }
}

void DiffusionAgent::ForwardMessage(Message *msg, Filter_Entry *dst)
{
  RedirectMessage *originalHdr;
  NRAttribute *originalAttr;
  int len;
  char *pos;
  struct hdr_diff *out_dfh;

  // Create an attribute with the original header
  originalHdr = new RedirectMessage;
  originalHdr->msg_type = msg->msg_type;
  originalHdr->source_port = msg->source_port;
  originalHdr->data_len = msg->data_len;
  originalHdr->num_attr = msg->num_attr;
  originalHdr->pkt_num = msg->pkt_num;
  originalHdr->rdm_id = msg->rdm_id;
  originalHdr->next_hop = msg->next_hop;
  originalHdr->last_hop = msg->last_hop;
  originalHdr->new_message = msg->new_message;
  originalHdr->handle = dst->handle;

  originalAttr = OriginalHdrAttr.make(NRAttribute::IS, (void *)originalHdr, sizeof(RedirectMessage));

  msg->msg_attr_vec->push_back(originalAttr);

  pos = (char *) &out_buffer;
  pos = pos + sizeof(struct hdr_diff);

  len = PackAttrs(msg->msg_attr_vec, pos);

  LAST_HOP(out_packet_dfh) = htonl(myId);
  NEXT_HOP(out_packet_dfh) = htonl(LOCALHOST_ADDR);
  NUM_ATTR(out_packet_dfh) = htons(msg->num_attr + 1);
  PKT_NUM(out_packet_dfh) = htonl(pkt_count);
  RDM_ID(out_packet_dfh) = htonl(rdm_id);
  SRC_PORT(out_packet_dfh) = htons(DIFFUSION_PORT);
  VERSION(out_packet_dfh) = DIFFUSION_VERSION;
  MSG_TYPE(out_packet_dfh) = REDIRECT;
  DATA_LEN(out_packet_dfh) = htons(len);

  // Send Packet
  SEND_TO(dst->skt_info, out_buffer);

  pkt_count++;
  delete originalHdr;
}

void DiffusionAgent::UpdateNeighbors(int id)
{
  NeighborList::iterator itr;
  Neighbor_Entry *newNeighbor;

  if (id == LOCALHOST_ADDR || id == myId)
    return;

  for (itr = my_neighbors.begin(); itr != my_neighbors.end(); ++itr){
    if ((*itr)->id == id)
      break;
  }

  if (itr == my_neighbors.end()){
    // This is a new neighbor
    newNeighbor = new Neighbor_Entry(id);
    my_neighbors.push_front(newNeighbor);
  }
  else{
    // Just update the neighbor timeout
    gettimeofday(&((*itr)->tmv), NULL);
  }
}

Filter_Entry * DiffusionAgent::FindFilter(int16_t handle, u_int16_t agent)
{
  FilterList::iterator itr;
  Filter_Entry *current;

  for (itr = filter_list.begin(); itr != filter_list.end(); ++itr){
    current = *itr;
    if (handle != current->handle || agent != current->agent)
      continue;

    // Found
    return current;
  }
  return NULL;
}

Filter_Entry * DiffusionAgent::DeleteFilter(int16_t handle, u_int16_t agent)
{
  FilterList::iterator itr = filter_list.begin();
  Filter_Entry *current = NULL;

  while (itr != filter_list.end()){
    current = *itr;
    if (handle == current->handle && agent == current->agent){
      filter_list.erase(itr);
      break;
    }
    current = NULL;
    itr++;
  }
  return current;
}

void DiffusionAgent::AddFilter(NRAttrVec *attrs, u_int16_t agent,
			       int16_t handle, u_int16_t priority)
{
  Filter_Entry *entry;

  entry = new Filter_Entry(handle, priority, agent);

#ifndef WIRED
  host_ent = GetThisHostEnt();
#else
  host_ent = GetThisHostEntWired();
#endif

  SKT_INFO(entry) = CreateSocket(host_ent, agent);

  if (SKT_INFO(entry) == NULL){
    fprintf(stderr, "Can't get a socket for the port %d.\n", agent);
    exit(-1);
  }

  // Copy the Attribute Vector
  entry->filterAttrs = CopyAttrs(attrs);

  // Add this filter to the filter list
  filter_list.push_back(entry);
}

FilterList::iterator DiffusionAgent::FindMatchingFilter(NRAttrVec *attrs,
							FilterList::iterator itr)
{
  Filter_Entry *entry;

  for (;itr != filter_list.end(); ++itr){
    entry = *itr;

    if (OneWayMatch(entry->filterAttrs, attrs)){
      // That's a match !
      break;
    }
  }
  return itr;
}

void DiffusionAgent::UpdateAgents(Message *msg)
{
  AgentList::iterator itr;
  Agent_Entry *entry;

  for (itr = agent_list.begin(); itr != agent_list.end(); ++itr){
    entry = *itr;
    if (entry->port == msg->source_port)
      return;
  }

  // Ok, it's a new agent
  entry = new Agent_Entry;
  entry->port = msg->source_port;

#ifndef WIRED
  host_ent = GetThisHostEnt();
#else
  host_ent = GetThisHostEntWired();
#endif

  entry->skt_info = CreateSocket(host_ent, msg->source_port);

  if (entry->skt_info == NULL){
    fprintf(stderr, "Fatal Error: DiffusionAgent::UpdateAgents can't create socket to port %d !\n", msg->source_port);
    exit(-1);
  }

  // Add new agent to the list
  agent_list.push_back(entry);
}

bool DiffusionAgent::RestoreOriginalHeader(Message *msg)
{
  NRAttrVec::iterator place = msg->msg_attr_vec->begin();
  NRSimpleAttribute<void *> *originalHeader = NULL;
  RedirectMessage *originalHdr;

  // Find original Header
  originalHeader = OriginalHdrAttr.find_from(msg->msg_attr_vec, place, &place);
  if (!originalHeader){
    fprintf(stderr, "Error: DiffusionAgent::ProcessControlMessage couldn't find the OriginalHdrAttr !\n");
    return false;
  }

  // Restore original Header
  originalHdr = (RedirectMessage *) originalHeader->getVal();

  msg->msg_type = originalHdr->msg_type;
  msg->source_port = originalHdr->source_port;
  msg->pkt_num = originalHdr->pkt_num;
  msg->rdm_id = originalHdr->rdm_id;
  msg->next_hop = originalHdr->next_hop;
  msg->last_hop = originalHdr->last_hop;
  msg->new_message = originalHdr->new_message;
  msg->num_attr = originalHdr->num_attr;
  msg->data_len = originalHdr->data_len;

  // Delete attribute from original set
  msg->msg_attr_vec->erase(place);
  delete originalHeader;

  return true;
}

FilterList * DiffusionAgent::GetFilterList(NRAttrVec *attrs)
{
  FilterList *my_list = new FilterList;
  FilterList::iterator itr, myItr;
  Filter_Entry *entry, *listEntry;

  // We need to come up with a list of filters to call
  // F1 will be called before F2 if F1->priority > F2->priority
  // If F1 and F2 have the same priority, F1 will be called
  // before F2 is F1->handle < F2->handle. If both handles
  // are equal, the one with the lower port number will be
  // called first.

  itr = FindMatchingFilter(attrs, filter_list.begin());

  while (itr != filter_list.end()){
    // We have a match
    entry = *itr;

    for (myItr = my_list->begin(); myItr != my_list->end(); ++myItr){
      listEntry = *myItr;

      // Figure out where to insert 
      if (entry->priority > listEntry->priority)
	break;

      if (entry->priority == listEntry->priority){
	if (entry->handle < listEntry->handle)
	  break;

	if (entry->handle == listEntry->handle)
	  if (entry->agent < listEntry->agent)
	    break;

      }
    }

    my_list->insert(myItr, entry);

    itr++;

    itr = FindMatchingFilter(attrs, itr);
  }
  return my_list;
}

void DiffusionAgent::ProcessMessage(Message *msg)
{
  // If message is from a new agent, we add it to our list.
  if (msg->last_hop == LOCALHOST_ADDR)
    UpdateAgents(msg);

  // First check if it's a control message
  if (msg->msg_type == CONTROL){
    ProcessControlMessage(msg);
    return;
  }

  // Now process the incoming message
  FilterList *my_list;
  FilterList::iterator itr;
  Filter_Entry *entry;

  my_list = GetFilterList(msg->msg_attr_vec);

  // Ok, we have a list of Filters to call. Send this message
  // to the first filter on this list
  if (my_list->size() > 0){
    itr = my_list->begin();
    entry = *itr;

    ForwardMessage(msg, entry);
    my_list->clear();
  }
  delete my_list;
}

void DiffusionAgent::ProcessControlMessage(Message *msg)
{
  NRSimpleAttribute<void *> *ctrlmsg = NULL;
  NRAttrVec::iterator place;
  ControlMessage *controlblob = NULL;
  FilterList *flist;
  FilterList::iterator listitr;
  Filter_Entry *entry;
  int command, param1, param2;
  u_int16_t priority, source_port;
  int16_t handle;
  bool filter_is_last = false;

  // Control messages should not come from other nodes
  if (msg->last_hop != LOCALHOST_ADDR){
    fprintf(stderr, "Error: Received control message from another node !\n");
    return;
  }

  // Find the control attribute
  place = msg->msg_attr_vec->begin();
  ctrlmsg = ControlMsgAttr.find_from(msg->msg_attr_vec, place, &place);

  if (!ctrlmsg){
    // Control message is invalid
    fprintf(stderr, "Error: Control message received is invalid !\n");
    return;
  }

  // Extract the control message info
  controlblob = (ControlMessage *) ctrlmsg->getVal();
  command = controlblob->command;
  param1 = controlblob->param1;
  param2 = controlblob->param2;

  // Filter API definitions
  //
  // command = ADD_FILTER
  // param1  = priority
  // param2  = handle
  // attrs   = other attrs specify the filter
  // 
  // Remarks: If this filter is already present for this module,
  //          we don't create a new one. A filter is identified
  //          by the handle and the originating agent. We output
  //          a warning message if the same handle is added twice
  //          and do nothing. However, if attrs and handle are
  //          the same, we just update priority.
  //
  //
  // command = REMOVE_FILTER
  // param1  = handle
  //
  // Remarks: Remove the filter identified by (agent, handle)
  //          If it's not found, a warning message is generated.
  //
  //
  // command = KEEPALIVE_FILTER
  // param1  = handle
  //
  // Remarks: Update the filter's timeout so it doesn't expire.
  //          If filter is not found, output a warning message.
  //
  //
  // command = SEND_MESSAGE
  // param1  = agent_id
  //
  // Remarks: Send message from a local App to another App ot
  //          a neighbor. If agent_id is zero, the packet goes
  //          out to the network. Otherwise, it goes to the
  //          agent_id located on this node.
  //
  //
  // command = SEND_TO_NEXT
  // param1  = handle
  //
  // Remarks: Send this message to the next application in the
  //          priority list. We have to assemble the list again
  //          and figure out the current agent's position on the
  //          list. Then, we send to the next guy.

  LogControlMessage(msg, command, param1, param2);

  // First we remove the control attribute from the message
  msg->msg_attr_vec->erase(place);
  delete ctrlmsg;

  switch(command){
  case ADD_FILTER:

    priority = param1;
    handle = param2;

    entry = FindFilter(handle, msg->source_port);

    if (entry){
      // Filter already present

      if (PerfectMatch(entry->filterAttrs, msg->msg_attr_vec)){
	// Attrs also match
	if (priority == entry->priority){
	  // Nothing to do !
	  fprintf(stderr, "Attempt to add filter %d, %d, %d twice !\n", msg->source_port, handle, priority);
	}
	else{
	  // Update the priority
	  fprintf(stderr, "Updated priority of filter %d, %d, %d to %d\n", msg->source_port, handle, entry->priority, priority);
	  entry->priority = priority;
	}

	break;
      }
    }

    // This is a new filter
    fprintf(stderr, "Adding filter %d, %d, %d\n", msg->source_port, handle, priority);
    AddFilter(msg->msg_attr_vec, msg->source_port, handle, priority);

    break;

  case REMOVE_FILTER:

    handle = param1;
    entry = DeleteFilter(handle, msg->source_port);
    if (entry){
      // Filter deleted
      fprintf(stderr, "Filter %d, %d, %d deleted.\n", entry->agent, entry->handle, entry->priority);
      delete entry;
    }
    else{
      fprintf(stderr, "Couldn't find filter to delete !\n");
    }

    break;

  case KEEPALIVE_FILTER:

    handle = param1;
    entry = FindFilter(handle, msg->source_port);
    if (entry){
      // Update filter timeout
      gettimeofday(&(entry->tmv), NULL);
      fprintf(stderr, "Filter %d, %d, %d refreshed.\n", entry->agent, entry->handle, entry->priority);
    }
    else{
      fprintf(stderr, "Couldn't find filter to update !\n");
    }

    break;

  case SEND_MESSAGE:

    if (!RestoreOriginalHeader(msg))
      break;

    // Send Message
    SendMessage(msg, param1);

    break;

  case SEND_TO_NEXT:

    handle = param1;
    source_port = msg->source_port;

    if (!RestoreOriginalHeader(msg))
      break;

    // Now process the incoming message
    flist = GetFilterList(msg->msg_attr_vec);

    // Find the filter after the 'current' filter on the list
    if (flist->size() > 0){
      for (listitr = flist->begin(); listitr != flist->end(); ++listitr){
	entry = *listitr;
	if (entry->agent == source_port &&
	    entry->handle == handle){
	  // This is the filter who sent us the info
	  listitr++;
	  if (listitr == flist->end())
	    filter_is_last = true;
	  break;
	}
      }

      if (!filter_is_last){
	if (listitr == flist->end())
	  listitr = flist->begin();
	entry = *listitr;

	ForwardMessage(msg, entry);

      }

      flist->clear();
    }

    delete flist;

    break;

  default:

    fprintf(stderr, "Error: Unknown control message received !\n");
    break;
  }
}

void DiffusionAgent::LogControlMessage(Message *msg, int command,
				       int param1, int param2)
{
  // Logs the incoming message
}

void DiffusionAgent::usage()
{
  fprintf(stderr, "Usage: diffusion [-d debug] [-f filename] [-t stoptime] [-v] [-h] [-s]\n\n");
  fprintf(stderr, "\t-d - Sets debug level (0-10)\n");
  fprintf(stderr, "\t-t - Schedule diffusion to exit after stoptime seconds\n");
  fprintf(stderr, "\t-f - Uses filename as the config file\n");
  fprintf(stderr, "\t-v - Prints diffusion version\n");
  fprintf(stderr, "\t-h - Prints this information\n");
  fprintf(stderr, "\t-s - Set time to 0.0\n");

  fprintf(stderr, "\n");

  exit(0);
}

DiffusionAgent::DiffusionAgent(int argc, char **argv)
{
  int opt;
  char *scadds_env;
  long stop_time;
  struct timeval tv;
  struct timezone tz;

  scadds_env = getenv("scadds_addr");
 
  if (scadds_env != NULL){
    myId = atoi(scadds_env);
  }
  else{
    fprintf(stderr, "Diffusion : scadds_addr not set. Using random id.\n");

    // Generate random ID
    do{

      srand(time(NULL));
      myId = rand();

    }
    while(myId == LOCALHOST_ADDR || myId == BROADCAST_ADDR);
  }

  // Parse command line options
  opterr = 0;
  config_file = NULL;
  debug_level = 1;
  stop_time = 0;

  while (1){
    opt = getopt(argc, argv, "f:hd:vt:s");

    switch(opt){

    case 's': 

      tv.tv_sec = 0;
      tv.tv_usec = 0;
      tz.tz_minuteswest = 0;
      tz.tz_dsttime = 0;
      if (settimeofday(&tv, &tz) != 0){
	fprintf(stderr, "Diffusion Error: failed to set time to 0.0\n");
	exit(0);
      }
      else{
	fprintf(stderr, "Diffusion: succeeded to set time to 0.0\n");
      }

      break;

    case 't':

      stop_time = atol(optarg);
      if (stop_time <= 0){
	fprintf(stderr, "Diffusion Error: stop time must be > 0\n");
	exit(0);
      }
      else{
	fprintf(stderr, "%s will stop after %ld seconds\n", PROGRAM, stop_time);
      }

      break;

    case 'h':

      usage();

      break;

    case 'v':

      fprintf(stderr, "\n%s %s\n", PROGRAM, RELEASE);
      exit(0);

      break;

    case 'd':

      debug_level = atoi(optarg);

      if (debug_level < 1 || debug_level > 10){
	fprintf(stderr, "Error: Debug level outside range or missing !\n");
	usage();
      }

      break;

    case 'f':

      if (!strncasecmp(optarg, "-", 1)){
	fprintf(stderr, "Error: Parameter is missing !\n");
	usage();
      }

      config_file = strdup(optarg);

      break;

    case '?':

      fprintf(stderr, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage();

      break;

    case ':':

      fprintf(stderr, "Parameter missing !\n");
      usage();

      break;
    }

    if (opt == -1)
      break;
  }

  if (!config_file)
    config_file = strdup(DEFAULT_CONFIG_FILE);

  // Initialize variables
  lon = 0.0;
  lat = 0.0;

#ifdef STATS
  stats = new DiffusionStats(myId);
#  ifndef WIRED
#     ifdef USE_RPC
  rpcstats = new RPCStats(myId);
#     endif
#  endif
#endif

  srand(time(NULL));
  pkt_count = rand();
  rdm_id = rand();

  Tcl_InitHashTable(&htable, 2);

  in_packet_dfh = HDR_DIFF(in_buffer);
  out_packet_dfh = HDR_DIFF(out_buffer);
  addrlen = SZ_ADDR_INET;

  // Initialize eventQueue
  eq = new eventQueue;
  eq->eq_new();

  // Add timers to the eventQueue
  eq->eq_addAfter(NEIGHBORS_TIMER, NULL, NEIGHBORS_DELAY);
  eq->eq_addAfter(FILTER_TIMER, NULL, FILTER_DELAY);

  if (stop_time > 0) 
    eq->eq_addAfter(STOP_TIMER, NULL, stop_time * 1000);

  gettimeofday(&tv, &tz);
  fprintf(stderr, "Diffusion : starting at time %ld:%ld\n", tv.tv_sec, tv.tv_usec);
  fprintf(stderr, "Diffusion : Node id = %d\n", myId);

  // Bind to the diffusion socket
#ifndef WIRED
  host_ent = GetThisHostEnt();

  in_sock = CreateSocket(host_ent, DIFFUSION_PORT);
  if (in_sock == NULL){
    fprintf(stderr, "Can't get a socket for the port %d.\n", DIFFUSION_PORT);
    exit(-1);
  }

  RegisterSocket(in_sock);
  
#ifndef __MOTE_NIC__
  if ((kernel_radio = open(RPC_DEVICE_FILE, O_RDWR)) < 0){
    fprintf(stderr, "Can't initialize the radiometrix radio !\n");
    exit(-1);
  }
  fprintf(stderr, "Using kernel-module driver through " RPC_DEVICE_FILE "\n");

#else
  if ((kernel_radio = open(MOTE_NIC_DEVICE_FILE, O_RDWR)) < 0){
    fprintf(stderr, "Can't initialize the mote-nic.\n");
    exit(-1);
  }
#endif

#else
  in_sock = OpenIncomingSocket(DIFFUSION_PORT);

  wired = new DiffusionWired(config_file);
  kernel_radio = -1;
#endif
}

Hash_Entry * DiffusionAgent::GetHash(unsigned int pkt_num,
				     unsigned int rdm_id)
{
  unsigned int key[2];

  key[0] = pkt_num;
  key[1] = rdm_id;

  Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(&htable, (char *)key);

  if (entryPtr == NULL)
    return NULL;

  return (Hash_Entry *)Tcl_GetHashValue(entryPtr);
}

void DiffusionAgent::PutHash(unsigned int pkt_num,
			     unsigned int rdm_id)
{
  Tcl_HashEntry *entryPtr;
  Hash_Entry    *hashPtr;
  HashList::iterator itr;
  unsigned int key[2];
  int newPtr;

  if (hash_list.size() == HASH_TABLE_MAX_SIZE){
    // Hash table reached maximum size

    for (int i = 0; ((i < HASH_TABLE_REMOVE_AT_ONCE)
		     && (hash_list.size() > 0)); i++){
      itr = hash_list.begin();
      entryPtr = *itr;
      hashPtr = (Hash_Entry *) Tcl_GetHashValue(entryPtr);
      delete hashPtr;
      hash_list.erase(itr);
      Tcl_DeleteHashEntry(entryPtr);
    }
  }

  key[0] = pkt_num;
  key[1] = rdm_id;

  entryPtr = Tcl_CreateHashEntry(&htable, (char *)key, &newPtr);

  if (newPtr == 0){
    fprintf(stderr, "Key already exists in hash !\n");
    return;
  }

  hashPtr = new Hash_Entry;

  Tcl_SetHashValue(entryPtr, hashPtr);

  hash_list.push_back(entryPtr);
}

void DiffusionAgent::dvi_send(caddr_t data, int length)
{
#ifndef WIRED
  write(kernel_radio, data, length);
#else
  wired->send(data, length);
#endif
}

void DiffusionAgent::recv()
{
  Message *rcv_message = NULL;
  int8_t version, msg_type;
  u_int16_t data_len, num_attr, source_port;
  int32_t rdm_id, pkt_num, next_hop, last_hop;   

  // Read header
  version = VERSION(in_packet_dfh);
  msg_type = MSG_TYPE(in_packet_dfh);
  source_port = ntohs(SRC_PORT(in_packet_dfh));
  pkt_num = ntohl(PKT_NUM(in_packet_dfh));
  rdm_id = ntohl(RDM_ID(in_packet_dfh));
  num_attr = ntohs(NUM_ATTR(in_packet_dfh));
  next_hop = ntohl(NEXT_HOP(in_packet_dfh));
  last_hop = ntohl(LAST_HOP(in_packet_dfh));
  data_len = ntohs(DATA_LEN(in_packet_dfh));

  // Check version
  if (version != DIFFUSION_VERSION)
    return;

  // Check for ID conflict
  if (last_hop == myId){
    fprintf(stderr, "Error: A diffusion ID conflict has been detected !\n");
    exit(0);
  }

  // Address filtering
  if ((next_hop != BROADCAST_ADDR) &&
      (next_hop != LOCALHOST_ADDR) &&
      (next_hop != myId))
    return;

  // Packet is good, create a message
  rcv_message = new Message(version, msg_type, source_port, data_len,
			    num_attr, pkt_num, rdm_id, next_hop, last_hop);

  int total_size = sizeof(struct hdr_diff) + data_len;

  rcv_message->msg = new int [1 + (total_size / sizeof(int))];
  if (rcv_message->msg == NULL){
    fprintf(stderr, "Can't allocate memory. Duplicate Packet fails.\n");
    exit(-1);
  }

  bzero(rcv_message->msg, total_size);
  memcpy(rcv_message->msg, in_buffer, total_size);
  rcv_message->msg_size = total_size;

  // Read all attributes into the Message structure
  rcv_message->msg_attr_vec = UnpackAttrs(rcv_message->msg, num_attr);

  // Control Messages are unique and don't go to the hash
  if (rcv_message->msg_type != CONTROL){
    // Hash table keeps info about packets
    Tcl_HashEntry *entryPtr;
    unsigned int key[2];

    key[0] = pkt_num;
    key[1] = rdm_id;
    entryPtr = Tcl_FindHashEntry(&htable, (char *) key);

    if (entryPtr != NULL){
      fprintf(stderr, "Received old message !\n");
      rcv_message->new_message = 0;
    }
    else{
      // Add message to the hash table
      PutHash(key[0], key[1]);
      rcv_message->new_message = 1;
    }
  }

#ifdef STATS
  stats->logIncomingMessage(rcv_message);
#endif

  // Now, we look to the current filters to find a match
  ProcessMessage(rcv_message);

  // Don't forget to message when we're done
  delete rcv_message;
}

void DiffusionAgent::add_to_set(int socket, fd_set *set, int *max)
{
  FD_SET(socket, set);

  if (socket > *max)
    *max = socket;
}

void DiffusionAgent::TimetoStop()
{
  FILE *outfile = NULL;

  if (debug_level > 5){
    outfile = fopen("/tmp/diffusion.out", "w");
    if (outfile == NULL){
      printf("Diffusion Error: Can't create /tmp/diffusion.out\n");
      return;
    }
  }

#ifdef STATS
  stats->printStats(stdout);
  if (outfile)
    stats->printStats(outfile);
#  ifndef WIRED
#     ifdef USE_RPC
  rpcstats->printStats(stdout);
  if (outfile)
    rpcstats->printStats(outfile);
#     endif
#  endif
#endif

  if (outfile)
    fclose(outfile);
}

void DiffusionAgent::run()
{
  fd_set  fds;
  int     status, max_sock;
  struct timeval *tv;

  // Main Select Loop
  while (1){

    // Wait for incoming packets
    FD_ZERO(&fds);
    max_sock = 0;
    add_to_set(in_sock->skt_fd, &fds, &max_sock);

#ifndef WIRED
    add_to_set(kernel_radio, &fds, &max_sock);
#endif

    // Figure out how much to wait
    tv = eq->eq_nextTimer();

    fflush(stdout);
    status = select(max_sock+1, &fds, NULL, NULL, tv);

    if ((status == 0) || (eq->eq_topInPast())){
      // We got a timeout
      event *e = eq->eq_pop();

      // Timeouts
      switch (e->type){

      case NEIGHBORS_TIMER:

	NeighborsTimeOut();
	free(e);

	break;

      case FILTER_TIMER:

	FilterTimeOut();
	free(e);

	break;

      case STOP_TIMER:

	free(e);
	TimetoStop();
	exit(0);

	break;
      }
    }

    // Check for new packets
    if (status > 0){

      // Process incoming packets from diffusion port
      if (FD_ISSET(in_sock->skt_fd, &fds)){
	RECV_FROM(in_sock, in_buffer, addrlen);
	recv();
      }

#ifndef WIRED
      // Process incoming packets from kernel-module radio
      if (FD_ISSET(kernel_radio, &fds)){
	if (read(kernel_radio, in_buffer, (MAX_PKT_SIZE/sizeof(int))) > 0)
	  recv();
      }
#endif
    }

    // This should not happen
    if (status < 0)
      fprintf(stderr, "Select returned %d\n", status);
  }
}

Packet DupPacket(Packet pkt)
{
  int p_size;
  struct hdr_diff *dfh = HDR_DIFF(pkt);

  p_size = sizeof(struct hdr_diff) + ntohs(DATA_LEN(dfh));

  Packet dup = new int [1 + (p_size / sizeof(int))];

  if (dup == NULL){
    fprintf(stderr, "Can't allocate memory. Duplicate Packet fails.\n");
    exit(-1);
  }

  memcpy(dup, pkt, sizeof(struct hdr_diff) + ntohs(DATA_LEN(dfh)));

  return dup;
}

void main(int argc, char **argv)
{
  agent = new DiffusionAgent(argc, argv);

  signal(SIGINT, signal_handler);

  agent->run();
}
